#!/usr/bin/perl

# long.pl -- determine whether a file contains long lines
# Written by Eddie Antonio Santos <easantos@ualberta.ca>.

#                 DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE 
#                     (based on http://sam.zoy.org/wtfpl/)
#                           Version 2, December 2004 
#       
#       Copyright (C) 2012 Eddie Antonio Santos <easantos@ualberta.ca>
#       
#       Everyone is permitted to copy and distribute verbatim or modified 
#       copies of this license document, and changing it is allowed as long 
#       as the name is changed. 
#       
#                 DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE 
#       TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION 
#       
#        0. You just DO WHAT THE FUCK YOU WANT TO. 
#      

use strict;
no strict "subs";
use warnings;

use Getopt::Long;
use Pod::Usage;

# Globals -- Yes; globals.
my $long_line = 80; # Default definition of "long"
my $tab_stop = 8; # Default tab stop -- the *only* tab stop.
my $mode = \&normalMode;
my $version = '0.60';

my $versionStr = <<EOF;
$0 $version
coded by eddieantonio in Perl
EOF

sub main {
    # Parse arguments.
    my $options = parseArgs();

    # Display help or version, if requested.
    printHelp()    if $options->{'help'};
    printVersion() if $options->{'version'};

    # Redefine how long a long line should be, if specified.
    $long_line = $options->{'length'}    if $options->{'length'};
    # Tabs are people too!
    $tab_stop = $options->{'tab-stop'} if $options->{'tab-stop'};

    # Determine which mode to use.
    $mode = \&normalMode  if $options->{'normal'};
    $mode = \&laconicMode if $options->{'laconic'};
    $mode = \&quietMode   if $options->{'quiet'};

    run();
}

sub parseArgs {
    my $options = {};
    my $errstr = "";

    # Enable bundling
    Getopt::Long::Configure (qw(default bundling no_auto_version));

    # Options are:
    GetOptions($options,
        # -h    help
        # -a    laconic
        # -l    redefine length
        # -q    quiet
        # -t    redefine tab length
        # -v    version
        'help|h',
        'laconic|a',
        'length|l=i',
        'normal',
        'quiet|q',
        'tab-stop|t=i',
        'version|v',
        #  and (defined $options->{'length'} and $options->{'length'} > 0
        #        or $errstr = "Length must be positive integer")
        #  and ($options->{'tab-length'} > 0
        #        or $errstr = "Tab length must be positive integer")
    )
        or pod2usage(-verbose => 1, -exitval => 2);

    return $options;
}

sub printHelp {
    pod2usage(-verbose => 1);
}

sub printVersion {
    print $versionStr;
    exit;
}

sub run {
    my $long_lines = 0;

    if (not @ARGV) {
        # No files are specified; use standard input.
        $mode->(STDIN, "stdin");

    } else {
        # Run mode() on each file.
        foreach my $filename (@ARGV) {

            my $input;

            if ($filename ne '-') {
                # Open the file; die if the file cannot be opened.
                open ($input, "<:encoding(UTF-8)", $filename)
                    or die "Could not open \"$filename\": $!";
            } else {
                # Open `stdin`; die if it can't be opened (What.).
                open($input, "<-:encoding(UTF-8)")
                    or die "Could not open stdin: $!";
                $filename = 'stdin';
            }

            # Use our length analyzer function.
            $mode->($input, $filename);

        }

    }

}

# Go through each line and print the line number if the line is long.
sub normalMode {
    my ($file, $filename) = @_;
    
    my $line_number = 1;

    # Iterate through each line.
    foreach my $line (<$file>) {
        chomp $line; 

        if (realLength($line) >= $long_line) {
            print("Line $line_number is long in $filename\n");
        }

        $line_number++;

    }

}

# Much like normal mode before it, but shorter and (perhaps) machine-parsable.
sub laconicMode {
    my ($file, $filename) = @_;

    my $line_number = 1;

    foreach my $line (<$file>) {
        chomp $line;

        # And the statement itself is shorter too!
        print "$filename\t$line_number\n" if realLength($line) >= $long_line;

        $line_number++;

    }

}

# Exit as soon as one file is determined to be long.
sub quietMode {
    my ($file, $filename) = @_;

    foreach my $line (<$file>) {
        chomp $line;
        exit 1 if realLength($line) >= $long_line;
    }

}

sub realLength {
    # Like length() but takes into account tab stops for a predefined number of
    # characters (default: 8).
    my $string = shift;

    my $len = 0;

    foreach my $character (split //, $string) {
        if ($character eq "\t") {
            # Get length without caluclating tab stop...
            my $estimate = $len + $tab_stop;
            # Chop off characters that are not aligned to tab stop
            $len += $tab_stop - ($estimate % $tab_stop)
        } else {
            $len++;

        }
    }

    return $len;
}

# Final statement in the program -- run the main function
main();

__END__

=encoding utf-8

=head1 NAME

B<long> - line length scrutinizer

=head1 SYNOPSIS

B<long>
[mode]
[B<--length|>B<-l> I<len>]
[B<--tab-stop|>B<-t> I<len>]
[F<file> F<...>]

B<long>
[B<--help>|B<-h>]
[B<--version>|B<-v>]

=head1 DESCRIPTION

B<long> determines if a file (or files) contains long lines.  If a
long line is found, the file name its line number is printed.  Input
is read from standard input if no file is specified or if the
special symbol B<-> is used.

What constitutes as a long line can be defined with the B<--length>
options; tab stop length can also be changed with the B<--tab-stop>
option.

There are three modes of operation for B<long>: Normal mode, laconic
mode and quiet mode. Normal mode, as its name suggests, is invoked
by default or by explicitly providing the B<--normal> argument. In
normal mode, each long line causes a line to be printed to C<stdout>
the form of:

    Line x in file is long.

Quiet mode, invoked by either B<-q> or B<--quiet> will not print
anything at all; instead B<long> long returns with a non-zero status
as soon as at one long line is found.

Laconic mode is much like normal mode, but shorter. It prints the
filename, a tab, and the line number. That is:

    filename	linenumber

This output is guaranteed across versions, therefore it is usable in
scripts. See L</EXAMPLES> for a sample script.

=head2 OPTIONS

=over 8

=item B<-h>, B<--help>

Prints extended usage (part of this documentation) and exits.

=item B<-a>, B<--laconic>

Laconic mode -- prints each long line in a concise, machine-parsable
manner. The output follows the form C<fileE<9>linenumber>. 
This is intended to be easy to parse with common utilities such as
B<cut>, B<sed> or B<awk>. See L</EXAMPLES> for a way to combine
standard UNIX tools with this mode.

=item B<-l> I<len>, B<--length> I<len>

Defines a long line to be at least I<len> characters long.
(Default: 80)

=item B<--normal>

Normal mode [default] -- print long lines in a human readable way.

=item B<-q>, B<--quiet>

Quiet mode -- prints nothing, but exits with a non-zero status if at
least one line is long.

=item B<-t> I<len>, B<--tab-stop> I<len>

Defines the length of a tab stop to be I<len> chacters long. It is
strongly suggested that this be left at its default of 8, lest you
wish to unleash the wrath of L<Linus
Torvalds|http://kerneltrap.org/mailarchive/git/2007/10/16/344968>.

=item B<-v>, B<--version>

Prints version information and exits.

=back

=head1 EXAMPLES

Standard usage is simple:

    long file.txt

If no argument is given, standard input is read:

    ps aux | long

One thing the author does commonly is to check whether all files in
a project are consistant to its style guidelines. Often, this means
that lines are limited to a certain length, say 80 characters. This
line 1) finds all files recursivly, 2) runs B<file> on them (via
B<xargs> which is often configured to print 'text' if a file looks
like text file.  3) This output is filtered by B<grep>, which 4) is
given to B<cut> which can then extract only the part before the
first colon (':').  5) This argument list is fed into B<long> via
xargs. 

    find -X . | xargs file | grep 'text' | cut -f 1 -d ':' \
        | xargs long --laconic --length 80


This can be automated in the project's Makefile as demonstrated:
    
    qa: test
    	-file $(INCLUDEDFILES) | grep 'text' | cut -s -f 1 -d ':' \
    	      | xargs long --length 80 --quiet
        

=head1 CREDITS

Written in 2012 by eddieantonio.

=cut
